﻿using System;
using System.Windows.Forms;
using System.Diagnostics;
using System.IO;
using System.Drawing;
using System.Linq;

//
// UnityLibraryCleaner
// A simple directory parser that looks for Unity Library sub-directories and deletes them.
// 
namespace UnityLibraryCleaner
{

    public partial class Form1 : Form
    {

        private readonly string[] DirectoriesToIgnore = { "$RECYCLE.BIN", "System Volume Information", ".git" };
        private readonly string libraryDirectoryName = "Library";
        private readonly string requiredSibling = "Assets";
        private readonly string metaFileFormat = ".meta";

        private string dataFilename = "data.txt";
        private string previouslyUsedParentDirectory = "";
        private string fileWarning = "WARNING: DO NOT MODIFY CONTENTS OF THIS FILE.";
        private bool askForConfirmation = true;

        private long cleanedBytesThisRun = 0;
        private int bytesPerMegabyte = 1000000;

        private int cleanedMetaFilesThisRun = 0;

        public Form1()
        {
            InitializeComponent();
        }

        private void Form1_Load(object sender, EventArgs e)
        {

            deleteLibrariesButton.Enabled = false;
            deleteMetaFilesButton.Enabled = false;

            loadPreviouslyUsedParentDirectory();

        }

        private void deleteLibrariesButton_Click(object sender, EventArgs e)
        {

            addOutputLine("Waiting on confirmation...");
            string path = folderBrowserDialog1.SelectedPath;

            DialogResult result = MessageBox.Show("Are you sure you want to recursively delete all '" + libraryDirectoryName + "' sub-folders of contained in '" + path + "'?", "Please Confirm", MessageBoxButtons.YesNo, MessageBoxIcon.Warning, MessageBoxDefaultButton.Button2);

            // Confirm for realsies
            if (result == DialogResult.Yes)
            {

                if (path != "")
                {

                    addOutputLine("Processing " + path + "...");
                    exitButton.Enabled = false;
                    deleteMetaFilesButton.Enabled = false;
                    cleanFromFolder(path);

                }

            }
            else
            {
                addOutputLine("ABORTED.", Color.OrangeRed);
            }

        }

        private void selectFolderButton_Click(object sender, EventArgs e)
        {

            DialogResult myResult = folderBrowserDialog1.ShowDialog();

            if(myResult == DialogResult.OK)
            {

                addOutputLine(myResult.ToString() + ": '" + folderBrowserDialog1.SelectedPath + "'");

                deleteLibrariesButton.Enabled = true;
                deleteMetaFilesButton.Enabled = true;
                refreshDirectoryLabel();

            }
            else
            {
                addOutputLine(myResult.ToString() + ": None");
            }
            

        }

        private void cleanFromFolder(string parentDirectory)
        {

            cleanedBytesThisRun = 0;

            if (Directory.Exists(parentDirectory))
            {

                try
                {

                    string[] subdirectoryEntries = Directory.GetDirectories(parentDirectory);

                    foreach (string subdirectory in subdirectoryEntries)
                    {

                        if (shouldReadDirectory(subdirectory))
                        {
                            addOutputLine("Searching sub-directory '" + subdirectory + "'");
                            searchDirectoryForLibrary(subdirectory);
                        }
                        else
                        {
                            addOutputLine("Ignoring sub-directory '" + subdirectory + "'");
                        }
                                                
                    }

                }
                catch (Exception e)
                {
                    displayExceptionToUser("Error trying to search directory '"+ parentDirectory + "': " + e.Message);
                }
                finally { }

            }
            else
            {
                addOutputLine("Directory '" + parentDirectory + "' does not exist, even though it was selected.");
            }

            savePreviouslyUsedParentDirectory(parentDirectory);

            addOutputLine("DONE! Cleaned up " + (cleanedBytesThisRun / bytesPerMegabyte) + " MB.", Color.LimeGreen);
            exitButton.Enabled = true;
            deleteMetaFilesButton.Enabled = true;

        }

        private void searchDirectoryForLibrary(string parentDirectory)
        {

            if (Directory.Exists(parentDirectory))
            {

                try
                {

                    string[] subdirectoryEntries = Directory.GetDirectories(parentDirectory);

                    foreach (string subdirectory in subdirectoryEntries)
                    {

                        if (shouldReadDirectory(subdirectory))
                        {

                            // If we found a libraryDirectoryName, we want to delete it and all its contents (including folders!)
                            if (getJustFolderName(subdirectory) == libraryDirectoryName)
                            {

                                try
                                {

                                    string siblingTestPath = subdirectory.Substring(0, subdirectory.LastIndexOf('\\')) + "\\" + requiredSibling;

                                    if (Directory.Exists(siblingTestPath))
                                    {

                                        long byteCount = getDirectorySizeInBytes(subdirectory);

                                        try
                                        {

                                            bool canDeleteThisDir = false;

                                            // Confirm if required
                                            if (askForConfirmation)
                                            {

                                                DialogResult result = MessageBox.Show("Are you sure you want delete '" + subdirectory + "' ("+ (byteCount / bytesPerMegabyte) + " MB) ?", "Confirm Deletion", MessageBoxButtons.YesNo, MessageBoxIcon.Warning, MessageBoxDefaultButton.Button2);
                                                canDeleteThisDir = (result == DialogResult.Yes);

                                            }
                                            else
                                            {
                                                canDeleteThisDir = true;
                                            }

                                            
                                            if (canDeleteThisDir)
                                            {

                                                addOutputLine("Deleting Library '" + subdirectory + "' (" + (byteCount / bytesPerMegabyte) + " MB)", Color.Aqua);
                                                Directory.Delete(subdirectory, true);
                                                cleanedBytesThisRun += byteCount;

                                            }
                                            else
                                            {
                                                addOutputLine("Skipping Library '" + subdirectory + "' (" + (byteCount / bytesPerMegabyte) + " MB)", Color.Aquamarine);
                                            }

                                                

                                        }
                                        catch (Exception e)
                                        {
                                            displayExceptionToUser("Cannot delete '" + subdirectory + "':" + e.Message);
                                        }
                                        finally { }

                                    }
                                    else
                                    {
                                        addOutputLine("Encountered Library '" + subdirectory + "' with no '"+requiredSibling+"' sibling. Ignorning...", Color.Orange);
                                    }

                                }
                                catch (Exception e)
                                {
                                    displayExceptionToUser("Error trying to delete directory'"+ subdirectory + "': " + e.Message);
                                }
                                finally { }

                               
                            }
                            else
                            {
                                searchDirectoryForLibrary(subdirectory);
                            }

                        }
                        else
                        {
                            Debug.WriteLine("Ignoring Sub-Dir='" + subdirectory + "'");
                        }

                    }

                }
                catch (Exception e)
                {
                    displayExceptionToUser("Error trying to search sub directory '"+ parentDirectory + "': " + e.Message);
                }
                finally { }

            }

        }

        private bool shouldReadDirectory(string dirName)
        {

            bool shouldRead = true;
            string justFolderName = getJustFolderName(dirName);

            foreach (string s in DirectoriesToIgnore)
            {

                if (s == justFolderName)
                {
                    shouldRead = false;
                    break;
                }

            }

            return shouldRead;

        }

        private string getJustFolderName(string fullPath)
        {
            return fullPath.Substring(fullPath.LastIndexOf('\\') + 1);
        }

        private void displayExceptionToUser(string msg)
        {
            //MessageBox.Show("Exception: " + msg, "Look, An Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
            addOutputLine("Exception: " + msg, Color.OrangeRed);
        }

        private void exitButton_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        private void addOutputLine(string msg, Color? c = null)
        {

            Color textColor = c.GetValueOrDefault(outputTextBox.ForeColor);
            outputTextBox.SelectionStart = outputTextBox.TextLength;
            outputTextBox.SelectionLength = 0;
            outputTextBox.SelectionColor = textColor;
            outputTextBox.AppendText(msg + "\r\n");
            outputTextBox.SelectionColor = outputTextBox.ForeColor;

            // Force scroll to bottom for really big outputs
            outputTextBox.SelectionStart = outputTextBox.TextLength;
            outputTextBox.ScrollToCaret();

        }

        private void loadPreviouslyUsedParentDirectory()
        {

            string dir = Directory.GetCurrentDirectory();
            string filePath = dir + "\\" + dataFilename;

            try
            {

                if (File.Exists(filePath))
                {

                    string[] lines = File.ReadAllLines(filePath);

                    if (lines.Length == 2)
                    {

                        previouslyUsedParentDirectory = lines[1];
                        addOutputLine("Loaded previously used directory '" + previouslyUsedParentDirectory + "' as default.");

                        folderBrowserDialog1.SelectedPath = previouslyUsedParentDirectory;
                        deleteLibrariesButton.Enabled = true;
                        deleteMetaFilesButton.Enabled = true;
                        refreshDirectoryLabel();

                    }
                    else if(lines.Length != 0)
                    {
                        addOutputLine("It seems like data file '" + filePath + "' was edited. Ignoring previously selected directory.");
                    }

                }
                else
                {

                    addOutputLine("Data file '"+filePath+"' didn't exist. Creating new one.");
                    string[] lines = { fileWarning, previouslyUsedParentDirectory };
                    File.WriteAllLines(filePath, lines);

                }

            }
            catch (Exception e)
            {
                displayExceptionToUser("Error trying to read data file '"+filePath+"': " + e.Message);
            }
            finally
            {
            }

        }

        private void savePreviouslyUsedParentDirectory(string path)
        {

            string dir = Directory.GetCurrentDirectory();
            string filePath = dir + "\\" + dataFilename;

            previouslyUsedParentDirectory = path;

            try
            {

                string[] lines = { fileWarning, previouslyUsedParentDirectory };
                File.WriteAllLines(filePath, lines);

                previouslyUsedParentDirectory = lines[1];
                addOutputLine("Saved '" + previouslyUsedParentDirectory + "' as default directory.");

            }
            catch (Exception e)
            {
                displayExceptionToUser("Error trying to save directory selection to data file '"+filePath+"':" + e.Message);
            }
            finally { }

        }

        private void refreshDirectoryLabel()
        {
            selectedDirectoryLabel.Text = "Selected Directory: '" + folderBrowserDialog1.SelectedPath + "'";
        }

        private static long getDirectorySizeInBytes(string path)
        {

            DirectoryInfo dir = new DirectoryInfo(path);
            return dir.EnumerateFiles("*.*", SearchOption.AllDirectories).Sum(file => file.Length);

        }

        private void confirmationPromptCheckboxChangeHandler(object sender, EventArgs e)
        {
            askForConfirmation = confirmationPromptCheckbox.Checked;
        }

        private void deleteMetaFilesButton_Click(object sender, EventArgs e)
        {


            addOutputLine("Waiting on confirmation...");
            string path = folderBrowserDialog1.SelectedPath;

            DialogResult result = MessageBox.Show("Are you sure you want to recursively delete all "+ metaFileFormat + " files contained in '" + path + "' and its sub-folders?", "Please Confirm", MessageBoxButtons.YesNo, MessageBoxIcon.Warning, MessageBoxDefaultButton.Button2);

            // Confirm deletion of meta files for real
            if (result == DialogResult.Yes)
            {

                if (path != "")
                {

                    addOutputLine("Processing " + path + "...");
                    exitButton.Enabled = false;
                    deleteLibrariesButton.Enabled = false;
                    cleanMetaFilesFromFolder(path);

                }

            }
            else
            {
                addOutputLine("ABORTED.", Color.OrangeRed);
            }

        }

        private void cleanMetaFilesFromFolder(string parentDirectory)
        {

            cleanedMetaFilesThisRun = 0;

            if (Directory.Exists(parentDirectory))
            {

                try
                {
                   
                    addOutputLine("Cleaning meta files from '" + parentDirectory + "'");
                    searchDirectoryForMetaFiles(parentDirectory);

                }
                catch (Exception e)
                {
                    displayExceptionToUser("Error trying to search directory '" + parentDirectory + "': " + e.Message);
                }
                finally { }

            }
            else
            {
                addOutputLine("Directory '" + parentDirectory + "' does not exist, even though it was selected.");
            }

            savePreviouslyUsedParentDirectory(parentDirectory);
            addOutputLine("DONE! Removed up " + cleanedMetaFilesThisRun + " "+ metaFileFormat + " files", Color.LimeGreen);
            exitButton.Enabled = true;
            deleteLibrariesButton.Enabled = true;

        }

        private void searchDirectoryForMetaFiles(string parentDirectory)
        {

            if (Directory.Exists(parentDirectory))
            {

                try
                {

                    // Delete the meta files here
                    string[] allMetaFiles = Directory.GetFiles(parentDirectory, "*" + metaFileFormat);

                    foreach (string s in allMetaFiles)
                    {

                        addOutputLine("Deleting " + metaFileFormat + " file '" + s + "'", Color.Aqua);
                        File.Delete(s);
                        cleanedMetaFilesThisRun += 1;

                    }

                    // Then search sub-directories and clean those too
                    string[] subdirectoryEntries = Directory.GetDirectories(parentDirectory);

                    foreach (string subdirectory in subdirectoryEntries)
                    {

                        if (shouldReadDirectory(subdirectory))
                        {
                            searchDirectoryForMetaFiles(subdirectory);
                        }
                        else
                        {
                            Debug.WriteLine("Ignoring Sub-Dir='" + subdirectory + "'");
                        }

                    }

                }
                catch (Exception e)
                {
                    displayExceptionToUser("Cannot delete " + metaFileFormat + " file in directory '" + parentDirectory + "':" + e.Message);
                }
                finally { }

            }

        }
        
    }

}
